/*
    MassChroQ : Mass Chromatogram Quantification
    Copyright (C) 2014  Olivier Langella <Olivier.Langella@moulon.inra.fr>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "masschroq_cli.h"
#include "mcqsession.h"

#include <QCommandLineParser>
#include <QDebug>
#include <odsstream/odsexception.h>
#include <pappsomspp/processing/uimonitor/uimonitortextpercent.h>

masschroqCli::masschroqCli(QObject *parent) : QObject(parent)
{
  // get the instance of the main application
  app = QCoreApplication::instance();
  // setup everything here
  // create any global objects
  // setup debug and warning mode
}

void
masschroqCli::windaube_exit()
{
  //#if (WIN32)
  //    mcqout() << "Press any key then enter to exit" << endl;
  //    getchar();
  //#endif
}

// 10ms after the application starts this method will run
// all QT messaging is running at this point so threads, signals and slots
// will all work as expected.
void
masschroqCli::run()
{

  // QCoreApplication * app(this);
  // Add your main code here
  qDebug() << "MainClass.Run is executing";

  const QDateTime dt_begin = QDateTime::currentDateTime();
  const QString masschroq_dir_path(QCoreApplication::applicationDirPath());

  QCommandLineParser parser;
  parser.setApplicationDescription(
    "Perform Mass Chromatogram Quantification as indicated in the XML input "
    "FILE. For additional information, see the MassChroQ Homepage : "
    "http://pappso.inra.fr/bioinfo/masschroq/");
  parser.addHelpOption();
  parser.addVersionOption();
  parser.addPositionalArgument("MassChroqML",
                               tr("MassChroQ XML input file to execute."));
  QCommandLineOption validateOption(
    QStringList() << "validate",
    tr("check xml input file validity without running MassChroQ"));
  parser.addOption(validateOption);

  //    out << "  -c, --cpus integer\t" << endl;
  QCommandLineOption cpusOption(QStringList() << "c"
                                              << "cpus",
                                tr("sets the number of cpus you want to use"),
                                tr("number"));
  parser.addOption(cpusOption);
  //        out << "  -t, --tmpdir DIRECTORY\tuse DIRECTORY as temporary working
  //        direcory for masschroq : temporary files generated by masschroq
  //        during execution time (one file or several slices at a time per xml
  //        data file being analysed) will go there. By default DIRECTORY is the
  //        current working directory of masschroq. Use this option if you want
  //        to set another one." << endl;

  QCommandLineOption tmpOption(
    QStringList() << "t"
                  << "tmpdir",
    tr("use DIRECTORY as temporary working direcory for masschroq : "
       "temporary "
       "files generated by masschroq during execution time (one file or "
       "several slices at a time per xml data file being analysed) will go "
       "there. By default DIRECTORY is the current working directory of "
       "masschroq. Use this option if you want to set another one."),
    tr("directory"));
  parser.addOption(tmpOption);

  QCommandLineOption ondiskOption(
    QStringList() << "ondisk",
    tr("Use the temporary directory to store detected peaks on disk. "
       "MassChroQ "
       "requires less memory to run on huge datasets, but is a little bit "
       "slower due to disk performance"));
  parser.addOption(ondiskOption);

  // out << "  -p, --parse-peptides\t\tperform peptide text file parsing only,
  // no quantification. If the identified peptides are given to masschroq via
  // peptide text files (defined in FILE), he parses them, creates a new file
  // called parsed-peptides_FILE which contains the FILE content (minus the
  // lines relative to thr peptide text files to parse) plus the parsed peptides
  // information. By default masschroq continues analysis and quantification on
  // this new file. Use this option if you want masschroq only to parse the
  // petide files but not to continue quantification on them. This allows you to
  // check the parsed peptides and to add trace information on specific peptides
  // before quantification." << endl;

  QCommandLineOption parseOption(
    QStringList() << "p"
                  << "parse-peptides",
    tr("perform peptide text file parsing only, no quantification. If the "
       "identified peptides are given to masschroq via peptide text files "
       "(defined in FILE), he parses them, creates a new file called "
       "parsed-peptides_FILE which contains the FILE content (minus the "
       "lines "
       "relative to thr peptide text files to parse) plus the parsed "
       "peptides "
       "information. By default masschroq continues analysis and "
       "quantification on this new file. Use this option if you want "
       "masschroq "
       "only to parse the petide files but not to continue quantification on "
       "them. This allows you to check the parsed peptides and to add trace "
       "information on specific peptides before quantification."));
  parser.addOption(parseOption);

  parser.process(*app);


  pappso::UiMonitorTextPercent ui_monitor(mcqout());
  
  ui_monitor.setTitle(QObject::tr("MassChroQ version %1").arg(MASSCHROQ_VERSION));

  // QStringList arguments = app->arguments();
  // QString fileName;
  /// by default, temporary files go in the temporary directory of the system
  //  QString tmpDirName = QDir::tempPath();

  QString tmpDirName = parser.value(tmpOption);
  if(tmpDirName.isEmpty())
    {
      /// by default, temporary files go in the current working directory
      tmpDirName = QDir::currentPath();
    }
  else
    {
    }

  QString cpusStr = parser.value(cpusOption);
  if(cpusStr.isEmpty())
    {
      // mcqout() << "masschroqml_validation cpusStr.isEmpty()" << endl;
      Quantificator::setCpuNumber(1);
      // mcqout() << "masschroqml_validation cpusStr.isEmpty() 2"<< endl;
    }
  else
    {
      Quantificator::setCpuNumber(cpusStr.toUInt());
    }
  // mcqout() << "masschroqml_validation == 1"<< endl;
  McqSession::getInstance().setTmpDir(tmpDirName);
  McqSession::getInstance().setUnassignedPeaksOnDisk(
    parser.isSet(ondiskOption));

  bool peptide_translation    = parser.isSet(parseOption);
  bool masschroqml_validation = parser.isSet(validateOption);
  // mcqout() << "masschroqml_validation == true" << masschroqml_validation;

  QStringList file_list = parser.positionalArguments();
  if(file_list.size() == 0)
    {
      mcqerr() << "MassChroQ : missing input filename." << Qt::endl;
      mcqerr() << "Try 'masschroq --help' for more information." << Qt::endl;
      windaube_exit();
      app->exit(1);
    }
  if(file_list.size() > 1)
    {
      mcqerr() << "MassChroQ : too many input filenames." << Qt::endl;
      mcqerr() << "Try 'masschroq --help' for more information." << Qt::endl;
      windaube_exit();
      app->exit(1);
    }

  QString mcq_filename = file_list[0];

  if(peptide_translation && masschroqml_validation)
    {
      mcqerr() << "MassChroQ : invalid use." << Qt::endl;
      mcqerr() << "Try 'masschroq --help' for more information." << Qt::endl;
      windaube_exit();
      app->exit(1);
    }

  // ReporterBase * logger = new ReporterConsole();
  // logger->write("Test HOOOOOO \n");
  // mcqout() << "masschroqml_validation == true" << masschroqml_validation;
  MassChroq mass_chroq_engine;

  try
    {
      // defined application directory
      mass_chroq_engine.setMasschroqDir(masschroq_dir_path);
      mass_chroq_engine.setXmlFilename(mcq_filename);

      if(masschroqml_validation == true)
        {
          // mcqout() << "masschroqml_validation == true";
          if(mass_chroq_engine.validateXmlFile(ui_monitor) == true)
            {
              // mcqout() << "MassChroqML file is VALID" << endl;
              ui_monitor.setTitle(QObject::tr("MassChroqML file is VALID"));
            }
          // you must call quit when complete or the program will stay in the
          // messaging loop
          // quit();
        }
      else
        {
          // mcqout() << "masschroqml_validation == false";
          qDebug() << "masschroqml_validation == false";

          if(tmpDirName != QDir::tempPath())
            {
              ui_monitor.setStatus(
                QObject::tr("MassChroQ : temporary working directory is : %1")
                  .arg(McqSession::getInstance().getTmpDirName()));
            }
          mass_chroq_engine.setBeginDateTime(dt_begin);

          if(peptide_translation)
            {
              mass_chroq_engine.runDomParser();
            }
          else
            {
              mass_chroq_engine.runXmlFile(ui_monitor);
            }

          const QDateTime dt_end = QDateTime::currentDateTime();

          const Duration dur =
            Utilities::getDurationFromDates(dt_begin, dt_end);


          ui_monitor.setStatus(
            QObject::tr("MassChroQ's execution time was : \n%1 days %2 hours "
                        "%3 minutes %4 seconds")
              .arg(Utilities::getDaysFromDuration(dur))
              .arg(Utilities::getHoursFromDuration(dur))
              .arg(Utilities::getMinutesFromDuration(dur))
              .arg(Utilities::getSecondsFromDuration(dur)));
        }
      windaube_exit();
      // you must call quit when complete or the program will stay in the
      // messaging loop
      quit();
      //    delete logger;
    }
  catch(OdsException &error)
    {
      mcqerr() << "MassChroQ encountered an error:"
               << Qt::endl;
      mcqerr() << error.qwhat() << Qt::endl;

      windaube_exit();
      exit(1);
      app->exit(1);
    }
  catch(pappso::PappsoException &error)
    {
      mcqerr() << "MassChroQ encountered an error:"
               << Qt::endl;
      mcqerr() << error.qwhat() << Qt::endl;

      windaube_exit();
      exit(1);
      app->exit(1);
    }
  qDebug() << "before final quit()";

  // you must call quit when complete or the program will stay in the
  // messaging loop
  quit();
}

// call this routine to quit the application
void
masschroqCli::quit()
{
  // you can do some cleanup here
  // then do emit finished to signal CoreApplication to quit
  emit finished();
}

// shortly after quit is called the CoreApplication will signal this routine
// this is a good place to delete any objects that were created in the
// constructor and/or to stop any threads
void
masschroqCli::aboutToQuitApp()
{
  // stop threads
  // sleep(1);   // wait for threads to stop.
  // delete any objects
}
